﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel.Channels;
using System.ServiceModel;
using System.ServiceModel.Discovery;
using System.ServiceModel.Description;
using System.Xml;
using System.Text.RegularExpressions;

namespace DynamicServiceFactory
{
    public static class DynamicChannelFactory
    {
        public const int MAXRETRIES = 1;
        // this will only happen once during te application,or more often when something goes wrong
        private static ServiceLocator.IServiceLocator _locatorProxy=CreateLocatorProxy();

        public static TContract CreateChannel<TContract>()
        {
            ServiceEndpoint se = GetFullEndpointMetaData<TContract>();
            if (se !=null)
            {
                return ChannelFactory<TContract>.CreateChannel(se.Binding, se.Address);    
            }
            return default(TContract);
            
        }

        public static ServiceEndpoint GetFullEndpointMetaData<TContract>()
        {
            return GetFullEndpointMetaData(typeof(TContract));
        }
        public static ServiceEndpoint GetFullEndpointMetaData(Type contract)
        {
            // next we need an addres, we ask it from the locatorservice, there services will have announced themselves
            string address = GetAddress(contract);
            if (!string.IsNullOrWhiteSpace(address))
            {
                Match m = Regex.Match(address, @"(.+\.svc)");
                address = m.Groups[0].Value + "?wsdl";
                //with the address we need to get Full endpoint information, including MetaData about the service such as the binding.
                ServiceEndpointCollection endpoints = MetadataResolver.Resolve(contract, new Uri(address), MetadataExchangeClientMode.HttpGet);
                //the resolve filtered endpoints based on contract type so we expect only one endpoint, maybe later we can build 
                //a preference for a certain binding?
                return endpoints[0]; 
            }
            return null;
        }
        private static string GetAddress(Type contract)
        {
            bool error = false;
            int nrRetries = 0;
            string addres = null;
            ContractDescription cd = ContractDescription.GetContract(contract);
            do
            {
                try
                {

                    addres = _locatorProxy.GetAddress(new XmlQualifiedName(cd.Name, cd.Namespace));
                    error = false;
                }
                catch (FaultException)
                {
                    // service was not found
                    return null;
                }
                catch (CommunicationException)
                {
                    // something infra structural went wrong, maybe the locator service moved? Lets discover it again and try again
                    CreateLocatorProxy();
                    error = true;
                    nrRetries++;
                }
                
            } while (error && nrRetries < MAXRETRIES);
            return addres;
        }

        /// <summary>
        /// Finding via udp is still relatively slow, this will happen only the first time, if there is a well known address of this service
        /// its better to use that instead of locating it dynamically, its also better to define a well known binding for its endpoint.
        /// </summary>
        /// <returns></returns>
        private static ServiceLocator.IServiceLocator CreateLocatorProxy()
        {
            Type locatorContract = typeof(ServiceLocator.IServiceLocator);
            DiscoveryClient dc = new DiscoveryClient(new UdpDiscoveryEndpoint());
            FindCriteria criteria = FindCriteria.CreateMetadataExchangeEndpointCriteria(locatorContract);
            //dont forget to set the max results to one, this is a huge performance improvement
            criteria.MaxResults = 1;
            FindResponse response = dc.Find(criteria);
            if (response.Endpoints.Count > 0)
            {
                var metaExchangeEndpointData = response.Endpoints[0];
                ServiceEndpointCollection endpoints = MetadataResolver.Resolve(locatorContract, metaExchangeEndpointData.Address);
                if (endpoints.Count > 0)
                {
                    // The Locator service only has one endpoint which support locating
                    return new ServiceLocator.ServiceLocatorClient(endpoints[0].Binding, endpoints[0].Address);
                }
            }
            return null;
        }
    }
}
